package ch.usi.inf.sape.tracer.analyzer;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.rmi.Naming;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import ch.usi.inf.sape.tracer.analyzer.RuntimeEnv;
import ch.usi.inf.sape.tracer.analyzer.abstractops.SourceLineOperation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.ANEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.ARRAYLENGTH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.ATHROW_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BIPUSH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BasicBlockEntry_PseudoOperation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Binary_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent.Condition;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.CHECKCAST_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.CatchBlock_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X1_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X1_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_3_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP2_X2_4_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_X1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_X2_1_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.DUP_X2_2_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Division_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.GETFIELD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.GETSTATIC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.GOTO_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IF_ACMP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IF_ICMP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IF_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.IINC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INSTANCEOF_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKEARGS_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKERETURNING_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.LOOKUPSWITCH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MONITORENTER_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MONITOREXIT_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MULTIANEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.NEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.NEW_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.PUTFIELD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.PUTSTATIC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Rem_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SIPUSH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SWAP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.TABLESWITCH_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.Unary_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XALOAD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XASTORE_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XCONST_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XLDC_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XLOAD_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XPOP_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XRETURN_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XSTORE_Operation;
import ch.usi.inf.sape.tracer.analyzer.locations.HeapSpace;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo.Type;

/**
 * The Analyzer is called from the instrumentation produced by the Transformer.
 * 
 * This Analyzer has one public static method for each bytecode instruction.
 * It has two methods for the INVOKE* instructions: 
 * INVOKE*_calling is called before the INVOKE instruction,
 * INVOKE*_returning is called after.
 * It also has a methodEntry() method, called at the entry to every method.
 * This way, each method call is represented by four calls to the analyzer:
 * - INVOKE*_calling()
 * - methodEntry() 
 * - *RETURN()
 * - INVOKE_returning()
 * This allows to tie together the flow of arguments from caller to callee,
 * and the flow of the return value from the callee back to the caller.
 */
public class Analyzer {
	
	public static final boolean DO_ABSTRACTION = true; 
	public static final boolean REMOTELY_AVAILABLE = false;
	public static final boolean DO_LOG = false;
	private static Stack<Integer> bbs = new Stack<>();
	
	
	
	//TODO replace this with a variable which is configurable
	//through the build file
	public static boolean includePEIs = false;
	
	/**
	 * This field is used in control dependency analysis to distinguish 
	 * among different inovcations.
	 */
	private static int nextInvocationId = -1;
	
	static{
		//This is initialized to -1 because nobody invokes the main method
		bbs.push(nextInvocationId --);
	}
	
	/**
	 * @param name
	 * @param maxLocalArraySize
	 * @param localDetails
	 * @param maxStack
	 * @param argsCount
	 * @param values
	 * @param isStatic
	 * @param sourceFile
	 */
	public static void methodEntry(final String name, 
									final int maxLocalArraySize, 
									final String localDetails, 
									final int maxStack, 
									final int argsCount, 
									final Object[] values, 
									final boolean isStatic,
									final String sourceFile,
									final Object receiver) {
		log("Method entry: "+name, false);
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;
		
		
		ThreadCallStack tcs = ThreadCallStack.getCurrentThreadCallStack();
		
		/* If an already thrown exception is caught in an untraced method, 
		 * nobody sets the thrown flag to false and this can be misleading 
		 * in case the untraced method calls another traced method again. 
		 * So we explicitly set this flag to false at this point to remove 
		 * the possibility of such a case.
		 */
		tcs.setExceptionCaught();		
		
		INVOKEARGS_Operation op;
		if (tcs.isCallerInstrumented(name, argsCount, isStatic)){
			Activation currentMethodFrame = tcs.peekTopMostFrame();
			currentMethodFrame.setOwnerClass(Utils.findClassName(name));   		// set to the runtime type of the object
			op = currentMethodFrame.getProvider();
		}else{
			op = new INVOKEARGS_Operation(-1, name, argsCount, isStatic, false,"INVOKEUNKNOWN", HeapSpace.getOrDefineInstanceLocationMetaData(receiver));
			RuntimeEnv.getInstance().addNewBytecodeEvent(op);
		}
		
		tcs.peekTopMostFrame().setSourceFile(sourceFile);
		
		op.def(values, name, maxLocalArraySize, localDetails, maxStack, argsCount, isStatic);
		log(op.toString());
	}
	
	public static void ILOAD(int local, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLOAD_Operation op = new XLOAD_Operation(Bytecodes.ILOAD, local, MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LLOAD(int local, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLOAD_Operation op = new XLOAD_Operation(Bytecodes.LLOAD, local, MemoryLocationInfo.Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FLOAD(int local, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLOAD_Operation op = new XLOAD_Operation(Bytecodes.FLOAD, local, MemoryLocationInfo.Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DLOAD(int local, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLOAD_Operation op = new XLOAD_Operation(Bytecodes.DLOAD, local, MemoryLocationInfo.Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ALOAD(int local, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLOAD_Operation op = new XLOAD_Operation(Bytecodes.ALOAD, local, MemoryLocationInfo.Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ISTORE(int local, int instruction, int bcIndex) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XSTORE_Operation op = new XSTORE_Operation(Bytecodes.ISTORE, local, MemoryLocationInfo.Type.INT, instruction, bcIndex);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LSTORE(int local, int instruction, int bcIndex) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XSTORE_Operation op = new XSTORE_Operation(Bytecodes.LSTORE, local, MemoryLocationInfo.Type.LONG, instruction, bcIndex);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}	
	
	public static void FSTORE(int local, int instruction, int bcIndex) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XSTORE_Operation op = new XSTORE_Operation(Bytecodes.FSTORE, local, MemoryLocationInfo.Type.FLOAT, instruction, bcIndex);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DSTORE(int local, int instruction, int bcIndex) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XSTORE_Operation op = new XSTORE_Operation(Bytecodes.DSTORE, local, MemoryLocationInfo.Type.DOUBLE, instruction, bcIndex);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void ASTORE(int local, int instruction, int bcIndex) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XSTORE_Operation op = new XSTORE_Operation(Bytecodes.ASTORE, local, MemoryLocationInfo.Type.REFERENCE, instruction, bcIndex);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ACONST_NULL(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ACONST_NULL, Value.NULL, MemoryLocationInfo.Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_M1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_M1, new Value(-1), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_0(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_0, new Value(0), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_1, new Value(1), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_2, new Value(2), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_3(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_3, new Value(3), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_4(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_4, new Value(4), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ICONST_5(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.ICONST_5, new Value(5), MemoryLocationInfo.Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LCONST_0(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.LCONST_0, new Value(0L), MemoryLocationInfo.Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LCONST_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.LCONST_1, new Value(1L), MemoryLocationInfo.Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FCONST_0(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.FCONST_0, new Value(0F), MemoryLocationInfo.Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FCONST_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.FCONST_1, new Value(1F), MemoryLocationInfo.Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FCONST_2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.FCONST_2, new Value(2F), MemoryLocationInfo.Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void DCONST_0(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.DCONST_0, new Value(0.0), MemoryLocationInfo.Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DCONST_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XCONST_Operation op = new XCONST_Operation(Bytecodes.DCONST_1, new Value(1.0), MemoryLocationInfo.Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IALOAD(int[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = new Value(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.IALOAD, array, valueObject, index, accessVerified, Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LALOAD(long[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = new Value(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.LALOAD, array, valueObject, index, accessVerified, Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FALOAD(float[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = new Value(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.FALOAD, array, valueObject, index, accessVerified, Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DALOAD(double[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = new Value(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.DALOAD, array, valueObject, index, accessVerified, Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void AALOAD(Object[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = Utils.getValueForObject(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.AALOAD, array, valueObject, index, accessVerified, Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void BALOAD(Object obj, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		// BALOAD is used for both byte[] and boolean[]; 
		Value valueObject = null;
		boolean accessVerified = obj != null && index >=0 && index < Array.getLength(obj);
		XALOAD_Operation op;
		if (obj instanceof byte[]){
			byte[] array = (byte[]) obj;
			if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
				valueObject = new Value(array[index]);
			}
			op = new XALOAD_Operation(Bytecodes.BALOAD, obj, valueObject, index, accessVerified, Type.BYTE, instruction);
		}else{
			boolean[] array = (boolean[]) obj;
			if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
				valueObject = new Value(array[index]);
			}
			op = new XALOAD_Operation(Bytecodes.BALOAD, obj, valueObject, index, accessVerified, Type.BOOLEAN, instruction);
		}
		
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);	
	}
	
	public static void CALOAD(char[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = new Value(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.CALOAD, array, valueObject, index, accessVerified, Type.CHAR, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void SALOAD(short[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Value valueObject = null;
		boolean accessVerified = array != null && index >=0 && index < array.length;  
		if (accessVerified){    //emulate the behavior of runtime verifier before accessing an array element
			valueObject = new Value(array[index]);
		}
		XALOAD_Operation op = new XALOAD_Operation(Bytecodes.SALOAD, array, valueObject, index, accessVerified, Type.SHORT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IASTORE(int[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.IASTORE, array, index, accessVerified, Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LASTORE(long[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.LASTORE, array, index, accessVerified, Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FASTORE(float[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.FASTORE, array, index, accessVerified, Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DASTORE(double[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.DASTORE, array, index, accessVerified, Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void AASTORE(Object[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.AASTORE, array, index, accessVerified, Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void BASTORE(Object array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		//bastore could be used for both booleans and bytes
		
		boolean accessVerified;
		if (array != null && index < Array.getLength(array) && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		if (array instanceof byte[]){
			XASTORE_Operation op = new XASTORE_Operation(Bytecodes.BASTORE, array, index, accessVerified, Type.BYTE, instruction);
			RuntimeEnv.getInstance().addNewBytecodeEvent(op);
		}
		else{
			XASTORE_Operation op = new XASTORE_Operation(Bytecodes.BASTORE, array, index, accessVerified, Type.BOOLEAN, instruction);
			RuntimeEnv.getInstance().addNewBytecodeEvent(op);
		}
	}
	
	public static void CASTORE(char[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.CASTORE, array, index, accessVerified, Type.CHAR, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void SASTORE(short[] array, int index, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		boolean accessVerified;
		if (array != null && index < array.length && index >=0)
			accessVerified = true;
		else
			accessVerified = false;
		
		XASTORE_Operation op = new XASTORE_Operation(Bytecodes.SASTORE, array, index, accessVerified, Type.SHORT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}


	public static void POP(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XPOP_Operation op = new XPOP_Operation(Bytecodes.POP, 1, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void POP2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XPOP_Operation op = new XPOP_Operation(Bytecodes.POP2, 2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DUP(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP_Operation op = new DUP_Operation(Bytecodes.DUP, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	/**
	 * Form: value2, value1 -> value1, value2, value1
	 * value1 and value2 are values of category1 computational type
	 * 
	 * @param value1
	 * @param value2
	 * @param instruction
	 */
	public static void DUP_X1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP_X1_Operation op = new DUP_X1_Operation(Bytecodes.DUP_X1, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	/**
	 * Form 1: value3, value2, value1 -> value1, value3, value2, value1    (all values are of category 1 computational type)				  
	 * 
	 * @param value1
	 * @param value2
	 * @param instruction
	 */
	public static void DUP_X2_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP_X2_1_Operation op = new DUP_X2_1_Operation(Bytecodes.DUP_X2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	
	/**
	 * Form 2: value2, value1  ->  value1, value2, value1				  (value1 is a value of a category 1 computational type and value2 is a value of a category 2 computational type)
	 *  
	 * @param instruction
	 */
	public static void DUP_X2_2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP_X2_2_Operation op = new DUP_X2_2_Operation(Bytecodes.DUP_X2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	/**
	 * Form1: value2, value1   ->   value2, value1, value2, value1       where both value1 and value2 are values of a category 1 computational type               
	 * 
	 * @param value2
	 * @param value1
	 * @param instruction
	 */
	public static void DUP2_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_1_Operation op = new DUP2_1_Operation(Bytecodes.DUP2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	/**
	 * Form2: value  -> value, value        where value is a value of a category 2 computational type               
	 * 
	 * @param value2
	 * @param value1
	 * @param instruction
	 */
	public static void DUP2_2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_2_Operation op = new DUP2_2_Operation(Bytecodes.DUP2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	
	/**
	 * Form1: value3, value2, value1  ->  value2, value1, value3, value2, value1          
	 * where value1, value2, and value3 are all values of a category 1 computational type          
	 * 
	 * @param instruction
	 */
	public static void DUP2_X1_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_X1_1_Operation op = new DUP2_X1_1_Operation(Bytecodes.DUP2_X1, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	/**
	 * Form2: value2, value1    ->    value1, value2, value1     	
	 * where value1 is a value of a category 2 computational type and value2 is a value of a category 1 computational type
	 * 
	 * @param instruction
	 */
	public static void DUP2_X1_2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_X1_2_Operation op = new DUP2_X1_2_Operation(Bytecodes.DUP2_X1, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	
	/**
	 * Form1: value4, value3, value2, value1  ->   value2, value1, value4, value3, value2, value1      
	 * where value1, value2, value3, and value4 are all values of a category 1 computational type
	 * 
	 * @param instruction
	 */
	public static void DUP2_X2_1(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_X2_1_Operation op = new DUP2_X2_1_Operation(Bytecodes.DUP2_X2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	
	/**
	 * Form2: value3, value2, value1   ->    value1, value3, value2, value1    	
	 * where value1 is a value of a category 2 computational type and value2 and value3 are both values of a category 1 computational type
	 * 
	 * @param instruction
	 */
	public static void DUP2_X2_2(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_X2_2_Operation op = new DUP2_X2_2_Operation(Bytecodes.DUP2_X2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	/**
	 * Form3: value3, value2, value1   ->    value2, value1, value3, value2, value1       
	 * where value1 and value2 are both values of a category 1 computational type and value3 is a value of a category 2 computational type
	 * 
	 * @param instruction
	 */
	public static void DUP2_X2_3(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_X2_3_Operation op = new DUP2_X2_3_Operation(Bytecodes.DUP2_X2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	/**
	 * Form4: value2, value1    ->    value1, value2, value1		
	 * where value1 and value2 are both values of a category 2 computational type
	 * 
	 * @param instruction
	 */
	public static void DUP2_X2_4(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		DUP2_X2_4_Operation op = new DUP2_X2_4_Operation(Bytecodes.DUP2_X2, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	/**
	 * Form:  value2, value1  ->  value1, value2				
	 * value1 and value2 are both values of a category 1 computational type
	 * 
	 * @param instruction
	 */
	public static void SWAP(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		SWAP_Operation op = new SWAP_Operation(Bytecodes.SWAP, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IADD(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.IADD, "IADD", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LADD(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LADD, "LADD", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FADD(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FADD, "FADD", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);

	}
	
	public static void DADD(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DADD, "DADD", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ISUB(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.ISUB, "ISUB", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LSUB(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LSUB, "LSUB", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FSUB(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FSUB, "FSUB", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DSUB(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DSUB, "DSUB", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IMUL(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.IMUL, "IMUL", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LMUL(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LMUL, "LMUL", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FMUL(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FMUL, "FMUL", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DMUL(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DMUL, "DMUL", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IDIV(int instruction) {		//we hook before executing this instruction, because it may throw exception. So no value is provided!
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Division_Operation op = new Division_Operation(Bytecodes.IDIV, "IDIV", Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void LDIV(int instruction) {		//we hook before executing this instruction, because it may throw exception. So no value is provided!
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Division_Operation op = new Division_Operation(Bytecodes.LDIV, "LDIV", Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void FDIV(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FDIV, "FDIV", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DDIV(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DDIV, "DDIV", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IREM(int instruction) {		//we hook before executing this instruction, because it may throw exception. So no value is provided!
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Rem_Operation op = new Rem_Operation(Bytecodes.IREM, "IREM", Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LREM(int instruction) {		//we hook before executing this instruction, because it may throw exception. So no value is provided!
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Rem_Operation op = new Rem_Operation(Bytecodes.LREM, "LREM", Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FREM(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FREM, "FREM", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DREM(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DREM, "DREM", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void INEG(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.INEG, "INEG", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LNEG(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.LNEG, "LNEG", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void FNEG(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.FNEG, "FNEG", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void DNEG(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.DNEG, "DNEG", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ISHL(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.ISHL, "ISHL", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LSHL(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LSHL, "LSHL", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ISHR(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.ISHR, "ISHR", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LSHR(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LSHR, "LSHR", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IUSHR(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.IUSHR, "IUSHR", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LUSHR(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LUSHR, "LUSHR", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IAND(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.IAND, "IAND", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void LAND(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LAND, "LAND", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IOR(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.IOR, "IOR", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void LOR(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LOR, "LOR", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IXOR(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.IXOR, "IXOR", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LXOR(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LXOR, "LXOR", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void I2L(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.I2L, "I2L", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void I2F(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.I2F, "I2F", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void I2D(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.I2D, "I2D", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void L2I(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.L2I, "L2I", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void L2F(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.L2F, "L2F", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void L2D(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.L2D, "L2D", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void F2I(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.F2I, "F2I", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void F2L(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.F2L, "F2L", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void F2D(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.F2D, "F2D", new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void D2I(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.D2I, "D2I", new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void D2L(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.D2L, "D2L", new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void D2F(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.D2F, "D2F", new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void I2B(byte value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.I2B, "I2B", new Value(value), Type.BYTE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void I2C(char value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.I2C, "I2C", new Value(value), Type.CHAR, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void I2S(short value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Unary_Operation op = new Unary_Operation(Bytecodes.I2S, "I2S", new Value(value), Type.SHORT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LCMP(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.LCMP, "LCMP" , new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FCMPL(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FCMPL, "FCMPL" , new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void FCMPG(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.FCMPG, "FCMPG" , new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void DCMPL(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DCMPL, "DCMPL" , new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DCMPG(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		Binary_Operation op = new Binary_Operation(Bytecodes.DCMPG, "DCMPG" , new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	
	public static void IRETURN(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XRETURN_Operation op = new XRETURN_Operation(Bytecodes.IRETURN, Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LRETURN(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XRETURN_Operation op = new XRETURN_Operation(Bytecodes.LRETURN, Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void FRETURN(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XRETURN_Operation op = new XRETURN_Operation(Bytecodes.FRETURN, Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void DRETURN(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XRETURN_Operation op = new XRETURN_Operation(Bytecodes.DRETURN, Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ARETURN(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XRETURN_Operation op = new XRETURN_Operation(Bytecodes.ARETURN, Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void RETURN(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XRETURN_Operation op = new XRETURN_Operation(Bytecodes.RETURN, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void ARRAYLENGTH(Object array, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		ARRAYLENGTH_Operation op = new ARRAYLENGTH_Operation(Bytecodes.ARRAYLENGTH, array, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void ATHROW(Object exception, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		ATHROW_Operation op = new ATHROW_Operation(Bytecodes.ATHROW, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void MONITORENTER(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		MONITORENTER_Operation op = new MONITORENTER_Operation(Bytecodes.MONITORENTER, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void MONITOREXIT(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		MONITOREXIT_Operation op = new MONITOREXIT_Operation(Bytecodes.MONITOREXIT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void NEWARRAY_pre(int size, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = new NEWARRAY_Operation(Bytecodes.NEWARRAY, size, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op, false);
	}
	
	public static void NEWARRAY_post(boolean[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.BOOLEAN);
		
		log(op.toString(), DO_LOG);
	}

	public static void NEWARRAY_post(char[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.CHAR);
		
		log(op.toString(), DO_LOG);
	}

	public static void NEWARRAY_post(float[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.FLOAT);
		
		log(op.toString(), DO_LOG);
	}

	public static void NEWARRAY_post(double[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.DOUBLE);
		
		log(op.toString(), DO_LOG);
	}
	
	public static void NEWARRAY_post(byte[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.BYTE);
		
		log(op.toString(), DO_LOG);
	}
	
	public static void NEWARRAY_post(short[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.SHORT);
		
		log(op.toString(), DO_LOG);
	}
	
	public static void NEWARRAY_post(int[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.INT);
		
		log(op.toString(), DO_LOG);
	}

	public static void NEWARRAY_post(long[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEWARRAY_Operation op = (NEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, Type.LONG);
		
		log(op.toString(), DO_LOG);
	}

	public static void BIPUSH(int value, int instruction) { // we know it's a byte, should we use the type byte
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		BIPUSH_Operation op = new BIPUSH_Operation(Bytecodes.BIPUSH, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void SIPUSH(int value, int instruction) { // we know it's a short, should we use type short
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		SIPUSH_Operation op = new SIPUSH_Operation(Bytecodes.SIPUSH, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IFEQ(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFEQ, Condition.EQ, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFNE(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFNE, Condition.NE, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFLT(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFLT, Condition.LT, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFGE(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFGE, Condition.GE, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFGT(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFGT, Condition.GT, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFLE(int target, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFLE, Condition.LE, target, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IF_ICMPEQ(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ICMP_Operation op = new IF_ICMP_Operation(Bytecodes.IF_ICMPEQ, Condition.EQ, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IF_ICMPNE(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ICMP_Operation op = new IF_ICMP_Operation(Bytecodes.IF_ICMPNE, Condition.NE, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IF_ICMPLT(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ICMP_Operation op = new IF_ICMP_Operation(Bytecodes.IF_ICMPLT, Condition.LT, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IF_ICMPGE(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ICMP_Operation op = new IF_ICMP_Operation(Bytecodes.IF_ICMPGE, Condition.GE, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IF_ICMPGT(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ICMP_Operation op = new IF_ICMP_Operation(Bytecodes.IF_ICMPGT, Condition.GT, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IF_ICMPLE(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ICMP_Operation op = new IF_ICMP_Operation(Bytecodes.IF_ICMPLE, Condition.LE, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IF_ACMPEQ(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ACMP_Operation op = new IF_ACMP_Operation(Bytecodes.IF_ACMPEQ, Condition.EQ, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IF_ACMPNE(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_ACMP_Operation op = new IF_ACMP_Operation(Bytecodes.IF_ACMPNE, Condition.NE, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GOTO(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GOTO_Operation op = new GOTO_Operation(Bytecodes.GOTO, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFNULL(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFNULL, Condition.NULL, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void IFNONNULL(int offset, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IF_Operation op = new IF_Operation(Bytecodes.IFNONNULL, Condition.NONNULL, offset, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	
	public static void LDC_pre(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = new XLDC_Operation(Bytecodes.LDC, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LDC(Object reference) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = (XLDC_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(Utils.getValueForObject(reference));
	}
	
	public static void LDC(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = new XLDC_Operation(Bytecodes.LDC, new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LDC(long value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = new XLDC_Operation(Bytecodes.LDC, new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LDC(float value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = new XLDC_Operation(Bytecodes.LDC, new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LDC(double value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = new XLDC_Operation(Bytecodes.LDC, new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void LDC(String value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		XLDC_Operation op = new XLDC_Operation(Bytecodes.LDC, Utils.getValueForObject(value), Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void IINC(int newValue, int localIndex, int incr,  int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		IINC_Operation op = new IINC_Operation(Bytecodes.IINC, new Value(newValue), localIndex, incr, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void NEW(Object value, String desc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		NEW_Operation op = new NEW_Operation(Bytecodes.NEW, value, desc, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void ANEWARRAY_pre(int size, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		ANEWARRAY_Operation op = new ANEWARRAY_Operation(Bytecodes.ANEWARRAY, size, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op, false);
	}

	public static void ANEWARRAY_post(Object[] array) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		ANEWARRAY_Operation op = (ANEWARRAY_Operation)ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array);
		
		log(op.toString(), DO_LOG);
	}

	public static void CHECKCAST(int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		CHECKCAST_Operation op = new CHECKCAST_Operation(Bytecodes.CHECKCAST, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INSTANCEOF(int value, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INSTANCEOF_Operation op = new INSTANCEOF_Operation(Bytecodes.INSTANCEOF, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	//For getstatic operations, we hook after the instruction, because this instruction uses only one memory location (static field location) 
	//and this location is available only if there is not any problem regarding class/field resolution. If there is any resolution problem, we
	// cannot use reflection to find the read value before the actual execution of the bytecode. In other terms, no exception can happen as a 
	// result of runtime information passed to this instruction.
	
	public static void GETSTATIC_boolean(boolean value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.BOOLEAN, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_byte(byte value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.BYTE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_char(char value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.CHAR, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_double(double value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_float(float value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GETSTATIC_int(int value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_long(long value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_short(short value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, new Value(value), Type.SHORT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETSTATIC_object(Object value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETSTATIC_Operation op = new GETSTATIC_Operation(Bytecodes.GETSTATIC, klass, fieldName, Utils.getValueForObject(value), Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void PUTSTATIC(boolean value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(byte value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(char value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(double value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(float value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(int value, Class klass,  String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(long value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(short value, Class klass, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, new Value(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTSTATIC(Object value, Class klass,  String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTSTATIC_Operation op = new PUTSTATIC_Operation(Bytecodes.PUTSTATIC, klass, fieldName, Utils.getValueForObject(value), instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	private static Field findFieldFromHierarchy(Class klass, String fieldName){
		if (klass == null)
			return null;
		try {
//			System.out.println("Class :\t"+klass.getName()+"\tfield name :\t"+fieldName);
			return klass.getDeclaredField(fieldName);
		} catch (SecurityException | NoSuchFieldException e) {
			return findFieldFromHierarchy(klass.getSuperclass(), fieldName);
		}
		
	}
	
	public static void GETFIELD_boolean(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				boolean value = false;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getBoolean(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.BOOLEAN, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETFIELD_byte(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				byte value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getByte(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.BYTE, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GETFIELD_char(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				char value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getChar(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.CHAR, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETFIELD_double(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				double value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getDouble(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.DOUBLE, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GETFIELD_float(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				float value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getFloat(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.FLOAT, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GETFIELD_int(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				int value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getInt(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.INT, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void GETFIELD_long(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				long value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getLong(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.LONG, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GETFIELD_short(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				short value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.getShort(object);
				field.setAccessible(isAccessible);
				valueObject = new Value(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.SHORT, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void GETFIELD_object(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		GETFIELD_Operation op;
		if (object != null){
			Value valueObject;
			try {
				Object value;
				final Field field= findFieldFromHierarchy(object.getClass(), fieldName);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				value = field.get(object);
				field.setAccessible(isAccessible);
				valueObject = Utils.getValueForObject(value);
			} catch (IllegalArgumentException | IllegalAccessException | NullPointerException e) {
				e.printStackTrace();
				valueObject = Value.UNDEFINED;
			}
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, object, fieldName, valueObject, Type.REFERENCE, instruction);
		}else{
			op = new GETFIELD_Operation(Bytecodes.GETFIELD, instruction);
		}
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void PUTFIELD(Object object, String fieldName, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		PUTFIELD_Operation op = new PUTFIELD_Operation(Bytecodes.PUTFIELD, object, fieldName, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void INVOKE_calling(Object receiver, String opName, String nameDescClass, int argsCount, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

//		INVOKECALLING_Operation op = new INVOKECALLING_Operation(opName, nameDescClass, argsCount, instruction);
//		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
		boolean isStatic = opName.equals("INVOKESTATIC");
		
		INVOKEARGS_Operation op = new INVOKEARGS_Operation(instruction, 
				nameDescClass, 
				argsCount, 
				isStatic,
				true,
				opName,
				isStatic ? HeapSpace.getOrDefineInstanceLocationMetaData(receiver) :
								receiver.getClass().isArray() ? HeapSpace.getOrDefineArrayLocationMetaData(receiver) : 
										HeapSpace.getOrDefineInstanceLocationMetaData(receiver));  
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void INVOKE_returning(boolean returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.BOOLEAN, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INVOKE_returning(byte returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.BYTE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INVOKE_returning(char returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.CHAR, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INVOKE_returning(double returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.DOUBLE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INVOKE_returning(float returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.FLOAT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void INVOKE_returning(int returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.INT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void INVOKE_returning(long returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.LONG, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void INVOKE_returning(short returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, new Value(returnValue), opName, classNameDesc, Type.SHORT, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INVOKE_returning(Object returnValue, String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, Utils.getValueForObject(returnValue), opName, classNameDesc, Type.REFERENCE, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void INVOKE_returning(String opName, String classNameDesc, int instruction) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		INVOKERETURNING_Operation op = INVOKERETURNING_Operation.create(Bytecodes.INVOKERETURN, opName, classNameDesc, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}

	public static void MULTIANEWARRAY_pre(int dimensions, int instruction, String desc){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		MULTIANEWARRAY_Operation op = new MULTIANEWARRAY_Operation(Bytecodes.MULTIANEWARRAY, dimensions, instruction, Utils.findMultiArrayElementType(desc));
		RuntimeEnv.getInstance().addNewBytecodeEvent(op, false);
	}
	
	public static void MULTIANEWARRAY_post(Object[] array, int dimensions){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		MULTIANEWARRAY_Operation op = (MULTIANEWARRAY_Operation) ThreadCallStack.getCurrentThreadCallStack().getBytecodeOperationBeforeLast(0);
		op.def(array, dimensions);
		
		log(op.toString(), DO_LOG);
	}
	
	public static void LOOKUPSWITCH_INSN(int instruction){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		LOOKUPSWITCH_Operation op = new LOOKUPSWITCH_Operation(Bytecodes.LOOKUPSWITCH, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	public static void TABLESWITCH_INSN(int instruction){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		TABLESWITCH_Operation op = new TABLESWITCH_Operation(Bytecodes.TABLESWITCH, instruction);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	
	public static void CATCH_BLOCK(Throwable e){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;
		
		log("Ooops! Exception "+ e +" was thrown and caught!", false);
		
		//At this point, we are sure that there has been a BasicBlockEntryOperation executed,
		//so, we don't need to check if the topmost activation belongs to the current method.
		//It must belong to this method, because it is already fixed by BasicBlockEntryOp.
		BytecodeEvent op = new CatchBlock_Operation(Bytecodes.CATCH_BLOCK, -1, Utils.getValueForObject(e));
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
 	}
	
	public static void BBEntry(int bb, String name, int[] doms){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;
		
		BasicBlockEntry_PseudoOperation op = new BasicBlockEntry_PseudoOperation(Bytecodes.BASIC_BLOCK_ENTRY, SourceLineOperation.PSEUDO_VAR_OPERATION_LINE_NO, bb, name, doms);
		RuntimeEnv.getInstance().addNewBytecodeEvent(op);
	}
	
	/**
	 * This method no longer should be used as a way to
	 * determine the end of an execution.
	 */
	@Deprecated 
	public static void onExit(){
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		ThreadCallStack curr = ThreadCallStack.getCurrentThreadCallStack();
		if (!curr.isCallStackEmpty()){
			return;
		}
		
	}
	
//	public static void generateHtmlView(BytecodeSessionManager session){
//		
//		try
//        {
//            HTMLTableViewRMI obj = new HTMLTableViewRMIImpl(session);
//            // Bind this object instance to the name "HelloServer"
//            log("\nWaiting for incoming RMI requests from the web application ...", true);
//            Naming.rebind("AnalysisServer", obj);
//        }
//        catch (Exception e)
//        {
//            System.out.println("Analysis Server err: " + e.getMessage());
//            e.printStackTrace();
//        }
//	}
	
	/////////////////////////////////////////////////////////////
	/**
	 * An uncaught exception
	 * @param exception
	 */
	public static void handleException(Throwable exception) {
		if (!ThreadCallStack.getCurrentThreadCallStack().isTraceMode())
			return;

		log("Pop frame on exception!", false);
		exception.printStackTrace(System.out);
		
		ThreadCallStack tcs = ThreadCallStack.getCurrentThreadCallStack();
		Activation currFrame = tcs.peekTopMostFrame();
		if (!currFrame.isInitialized()){
			currFrame = tcs.popFrame();
			if (!tcs.isExceptionThrown()){		//This means an exception was thrown and uncaught in an untraced part of the code
				tcs.setUncaughtExceptionInUninstrumented();
			}
		}
		
		tcs.setExceptionThrown();
		
		tcs.popFrame();
		
		if (tcs.isCallStackEmpty()){
			onExit();    
		}
//		tcs.killInvoke();
//		final Activation callerFrame = tcs.peekTopMostFrame();
//		if (currFrame.getMethodName().equals("main") && currFrame.getDesc().equals("([Ljava/lang/String;)V")){
//			System.out.println("EXCEPTION HANDLER: Generating backward trace for the operands of the instruction throwing the exception: "+exception+"\n");
//			StackTraceElement[] elements = exception.getStackTrace();
//			for (StackTraceElement stackTraceElement : elements) {
//				System.out.println("EXCEPTION HANDLER:\t" + stackTraceElement);
//			}
//			BytecodeOperation op = tcs.getLastBytecodeOperation();
//			log (op.toString(), true);
//			for (int i = 0; i<op.uses.length; i++){
//				if (op.uses[i] != Binding.UNDEFINED_BINDING)
//					logException(op.uses[i].getProvider());
//			}
			
//			Trace.traceOrigin(trace.getTarget(), traces.size()-1);
//			ExcelSpreadSheetView.update(traces);
//			Trace.createAST();
//			Trace.createStatementAST();
//			System.out.println("EXCEPTION HANDLER:\t"+trace);

//			try {
//				DotGraph.drawExpGraph();
//				DotGraph.drawGraph();
//				DotGraph.drawFowardTraceGraph();
//			} catch (IOException e) {
//				e.printStackTrace();
//			}
//		}
	}
	
	private static void logException(BytecodeEvent op){
		log(op.toString(), true);
		for (int i = 0; i<op.uses.length; i++){
			if (op.uses[i] != Binding.UNDEFINED_BINDING || op.uses != Binding.NO_USE_BINDINGS)
				logException(op.uses[i].getProvider());
		}
	}
	
	
	/////////////////////////////////////////////////////////////
	public static void log(String message) {
		log(message, DO_LOG);
	}

	public static void log(String message, boolean write){
		if (write)
			System.out.println("TRACE:\t"+message);
	}
	

		

}

