package ch.usi.inf.sape.tracer.analyzer;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import ch.usi.inf.sape.tracer.analyzer.Binding;
import ch.usi.inf.sape.tracer.analyzer.BytecodeAnalyzer;
import ch.usi.inf.sape.tracer.analyzer.QueryFacade;
import ch.usi.inf.sape.tracer.analyzer.RuntimeEnv;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKEARGS_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.SlicedBytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.locations.PseudoVariableMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.StackSlotMemoryLocation;

/**
 * This class embodies the idea of the InfectoMeter,
 * a tool to compute the infection impact of each
 * line of an input program for a given input value.
 * The input must be wrapped as a unit test.
 * 
 * @author reza
 *
 */
public class InfectoAnalyzer implements BytecodeAnalyzer {
	
	/*For each source line no. keep track of its 
		maximum infection size as the number of memory 
		locations affected by any of the values defined 
		in this line*/ 
	private final Map<String, Map<Integer,Integer>> sourceLineMaxInfectionMap = new HashMap<>();
	

	@Override
	public void analyze(RuntimeEnv env) {
		System.out.println("InfectoMeter Launched!\nAnalysis may take quite a while ...");
		
		for (BytecodeEvent event : env.getAllBytecodeEvents()){

			if (QueryFacade.isInvokeInstruction(event) &&
					((INVOKEARGS_Operation) event).getCalleeFrame().getMethodName().startsWith("assert") && 
					((INVOKEARGS_Operation) event).getCalleeFrame().getOwnerClass().equals("org/junit/Assert")){


				//backward slice makes the search space more focused
				List<SlicedBytecodeEvent> slice = QueryFacade.findBackwardSliceForBytecode(event, true, true, env);

				/* for each event in backward slice and for each value
				defined by that event, find forward slice, count the 
				number of memory locations defined in the fs, update
				the value of infection size for the line no. corresponding
				to the seed bytecode.
				 */
				for (SlicedBytecodeEvent slicedBytecodeEvent : slice) {
					Set<BytecodeEvent> forwardSlice = QueryFacade.findDataForwardSliceForBytecode(slicedBytecodeEvent.getNonSlicedEvent(), true, true, env);
					int infectionSize = countInfectedMemoryLocations(forwardSlice);
					updateInfectionMap(slicedBytecodeEvent.getSourceFileName(), slicedBytecodeEvent.getSourceCodeLineNumber(), infectionSize);
				}
			}
		}
		
		InfectoJSONWriter writer = new InfectoJSONWriter(env.getAppAbsPath(), sourceLineMaxInfectionMap);
		
		try {
			writer.generateJsonFiles();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}

//		Set<String> sources = sourceLineMaxInfectionMap.keySet();
//		for (String source : sources) {
//			List<Entry<Integer,Integer>> entries = new ArrayList<>(sourceLineMaxInfectionMap.get(source).entrySet());
//			Collections.sort(entries, new Comparator<Entry<Integer,Integer>>(){
//				
//				@Override
//				public int compare(Entry<Integer, Integer> o1, Entry<Integer, Integer> o2) {
//					return o1.getKey() > o2.getKey() ? 1 : (o1.getKey() == o2.getKey() ? 0 : -1);
//				}
//				
//			});
//			
//			for (Entry<Integer, Integer> entry : entries) {
//				System.out.println(entry);
//				
//			}
//			
//		}
	}

	
	private int countInfectedMemoryLocations(Set<BytecodeEvent> forwardSlice){
		int count = 0;
		for (BytecodeEvent event : forwardSlice) {
			for (Binding<?> def : event.getDefs()){
				//temporary defined values are not considered
				if (def == null || 
						(def.getMemoryLocation() instanceof StackSlotMemoryLocation) ||
						(def.getMemoryLocation() instanceof PseudoVariableMemoryLocation))
					continue;

				count ++;
			}
		}
		return count;
	}

	private void updateInfectionMap(String source, int lineNo, int infectionSize){
		Map<Integer,Integer> lineImpactMap = sourceLineMaxInfectionMap.get(source);
		if (lineImpactMap == null){
			lineImpactMap = new HashMap<Integer,Integer>();
			sourceLineMaxInfectionMap.put(source, lineImpactMap);
		}
		
		Integer currInfectionSize = lineImpactMap.get(lineNo);
		if ( currInfectionSize == null ||
				infectionSize > currInfectionSize){
			lineImpactMap.put(lineNo, infectionSize);
		}
	}
}
